@@ -18,7 +18,7 @@
     License along with this library; if not, write to the Free
     Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 
-    $Id: util.c,v 1.113 1999/10/01 21:58:10 sopwith Exp $
+    $Id: util.c,v 1.126 2001/09/24 16:29:02 mmclouglin Exp $
 
 ***************************************************************************/
 #include <assert.h>
@@ -94,6 +94,8 @@ GHashTable *				__IDL_filename_hash;
 IDL_fileinfo *				__IDL_cur_fileinfo;
 GHashTable *				__IDL_structunion_ht;
 int					__IDL_inhibits;
+int					__IDL_typecodes_as_tok;
+int					__IDL_pidl;
 IDL_tree				__IDL_root;
 IDL_ns					__IDL_root_ns;
 int					__IDL_is_okay;
@@ -114,7 +116,7 @@ guint IDL_strcase_hash (gconstpointer v)
 	guint h = 0, g;
 
 	for (p = (char *) v; *p != '\0'; ++p) {
-		h = (h << 4) + isupper (*p) ? tolower (*p) : *p;
+		h = (h << 4) + isupper ((int)*p) ? tolower (*p) : *p;
 		if ((g = h & 0xf0000000)) {
 			h = h ^ (g >> 24);
 			h = h ^ g;
@@ -173,7 +175,7 @@ gint IDL_ident_cmp (gconstpointer a, gconstpointer b)
 
 const char *IDL_get_libver_string (void)
 {
-	return LIBIDL_LIBRARY_VERSION;
+	return LIBIDL_VERSION;
 }
 
 const char *IDL_get_IDLver_string (void)
@@ -190,6 +192,22 @@ static void IDL_tree_optimize (IDL_tree *p, IDL_ns ns)
 	IDL_tree_remove_empty_modules (p, ns);
 }
 
+static void IDL_parse_setup(unsigned long parse_flags, int max_msg_level) {
+	if (parse_flags & IDLF_XPIDL)
+		parse_flags |= IDLF_PROPERTIES;
+
+	__IDL_max_msg_level = max_msg_level;
+	__IDL_nerrors = __IDL_nwarnings = 0;
+	__IDL_inhibits = 0;
+	__IDL_typecodes_as_tok = (parse_flags & IDLF_TYPECODES) ? 1 : 0;
+	__IDL_pidl = (parse_flags & IDLF_XPIDL) ? 1 : 0;
+	__IDL_flags = parse_flags;
+	__IDL_flagsi = 0;
+	__IDL_is_parsing = TRUE;
+	__IDL_is_okay = TRUE;
+	__IDL_new_ident_comments = NULL;
+}
+
 int IDL_parse_filename (const char *filename, const char *cpp_args,
 			IDL_msg_callback msg_cb, IDL_tree *tree, IDL_ns *ns,
 			unsigned long parse_flags, int max_msg_level)
@@ -202,17 +220,26 @@ int IDL_parse_filename (const char *filename, const char *cpp_args,
 	char *cmd;
 	size_t cmd_len;
 #ifdef HAVE_CPP_PIPE_STDIN
-	char *fmt = CPP_PROGRAM " " CPP_NOSTDINC " - %s%s %s < \"%s\" 2>/dev/null";
+	char *fmt = CPP_PROGRAM " " CPP_NOSTDINC " - %s%s %s < \"%s\" %s";
 	char *wd = "", *dirend;
 #else
-	char *fmt = CPP_PROGRAM " " CPP_NOSTDINC " -I- -I%s %s \"%s\" 2>/dev/null";
+	char *fmt = CPP_PROGRAM " " CPP_NOSTDINC " -I- -I%s %s \"%s\" %s";
 	char *s, *tmpfilename;
 	char cwd[2048];
 	gchar *linkto;
 #endif
+	char *cpperrs = (parse_flags & IDLF_SHOW_CPP_ERRORS) 
+	  ? "" : "2>/dev/null";
 	GSList *slist;
 	int rv;
 
+#if 0 && defined(YYDEBUG)
+	{
+		extern int __IDL_debug;
+		__IDL_debug = 1;
+	}
+#endif
+
 	if (!filename ||
 	    !tree ||
 	    (tree == NULL && ns != NULL)) {
@@ -235,6 +262,7 @@ int IDL_parse_filename (const char *filename, const char *cpp_args,
 
 	cmd_len = (strlen (filename) + (*wd ? 2 : 0) + strlen (wd) +
 		  (cpp_args ? strlen (cpp_args) : 0) +
+		  strlen(cpperrs) +
 		  strlen (fmt) - 8 + 1);
 	cmd = g_malloc (cmd_len);
 	if (!cmd) {
@@ -243,7 +271,7 @@ int IDL_parse_filename (const char *filename, const char *cpp_args,
 	}
 
 	g_snprintf (cmd, cmd_len, fmt, *wd ? "-I" : "", wd,
-		    cpp_args ? cpp_args : "", filename);
+		    cpp_args ? cpp_args : "", filename, cpperrs);
 
 	if (dirend)
 		g_free (wd);
@@ -287,6 +315,7 @@ int IDL_parse_filename (const char *filename, const char *cpp_args,
 
 	cmd_len = (strlen (tmpfilename) + strlen (cwd) +
 		   (cpp_args ? strlen (cpp_args) : 0) +
+		   strlen(cpperrs) +
 		   strlen (fmt) - 6 + 1);
 	cmd = g_malloc (cmd_len);
 	if (!cmd) {
@@ -296,7 +325,7 @@ int IDL_parse_filename (const char *filename, const char *cpp_args,
 	}
 
 	g_snprintf (cmd, cmd_len, fmt,
-		    cwd, cpp_args ? cpp_args : "", tmpfilename);
+		    cwd, cpp_args ? cpp_args : "", tmpfilename, cpperrs);
 #endif
 
 #ifndef NO_POPEN
@@ -313,22 +342,12 @@ int IDL_parse_filename (const char *filename, const char *cpp_args,
 		return IDL_ERROR;
 	}
 
-	if (parse_flags & IDLF_XPIDL)
-		parse_flags |= IDLF_PROPERTIES;
+	IDL_parse_setup(parse_flags, max_msg_level);
 
-	__IDL_max_msg_level = max_msg_level;
-	__IDL_nerrors = __IDL_nwarnings = 0;
 	__IDL_in = input;
 	__IDL_msgcb = msg_cb;
-	__IDL_inhibits = 0;
-	__IDL_flags = parse_flags;
-	__IDL_flagsi = 0;
 	__IDL_root_ns = IDL_ns_new ();
-
-	__IDL_is_parsing = TRUE;
-	__IDL_is_okay = TRUE;
 	__IDL_lex_init ();
-	__IDL_new_ident_comments = NULL;
 
 	__IDL_real_filename = filename;
 #ifndef HAVE_CPP_PIPE_STDIN
@@ -414,23 +433,15 @@ int IDL_parse_filename_with_input (const char *filename,
 		return -1;
 	}
 
-	if (parse_flags & IDLF_XPIDL)
-		parse_flags |= IDLF_PROPERTIES;
 
-	__IDL_max_msg_level = max_msg_level;
-	__IDL_nerrors = __IDL_nwarnings = 0;
+	IDL_parse_setup(parse_flags, max_msg_level);
+
 	__IDL_msgcb = msg_cb;
-	__IDL_inhibits = 0;
-	__IDL_flags = parse_flags;
-	__IDL_flagsi = 0;
 	__IDL_root_ns = IDL_ns_new ();
 
-	__IDL_is_parsing = TRUE;
-	__IDL_is_okay = TRUE;
 	__IDL_lex_init ();
 	__IDL_inputcb = input_cb;
 	__IDL_inputcb_user_data = input_cb_user_data;
-	__IDL_new_ident_comments = NULL;
 
 	__IDL_real_filename = filename;
 #ifndef HAVE_CPP_PIPE_STDIN
@@ -1305,6 +1316,17 @@ IDL_tree IDL_codefrag_new (char *desc, GSList *lines)
 	return p;
 }
 
+IDL_tree IDL_srcfile_new (char *filename, int seenCnt, gboolean isTop, gboolean wasInhibit)
+{
+	IDL_tree p = IDL_node_new (IDLN_SRCFILE);
+	IDL_SRCFILE (p).filename = filename;
+	IDL_SRCFILE (p).seenCnt = seenCnt;
+	IDL_SRCFILE (p).isTop = isTop;
+	IDL_SRCFILE (p).wasInhibit = wasInhibit;
+
+	return p;
+}
+
 IDL_tree IDL_const_dcl_new (IDL_tree const_type, IDL_tree ident, IDL_tree const_exp)
 {
 	IDL_tree p = IDL_node_new (IDLN_CONST_DCL);
@@ -1544,6 +1566,7 @@ static void IDL_tree_walk_real (IDL_tree_func_data *tfd, IDLTreeWalkRealData *da
 
 	down_tfd.state = tfd->state;
 	down_tfd.up = tfd;
+	down_tfd.level = tfd->level + 1;
 
 	p = tfd->tree;
 
@@ -1565,6 +1588,7 @@ static void IDL_tree_walk_real (IDL_tree_func_data *tfd, IDLTreeWalkRealData *da
 	case IDLN_TYPE_INTEGER:
 	case IDLN_TYPE_CHAR:
 	case IDLN_CODEFRAG:
+	case IDLN_SRCFILE:
 		break;
 
 	case IDLN_LIST:
@@ -1581,8 +1605,10 @@ static void IDL_tree_walk_real (IDL_tree_func_data *tfd, IDLTreeWalkRealData *da
 	case IDLN_MEMBER:
 		down_tfd.tree = IDL_MEMBER (p).type_spec;
 		IDL_tree_walk_real (&down_tfd, data);
-		down_tfd.tree = IDL_MEMBER (p).dcls;
-		IDL_tree_walk_real (&down_tfd, data);
+		if ( (tfd->state->flags & IDL_WalkF_TypespecOnly)==0 ) {
+			down_tfd.tree = IDL_MEMBER (p).dcls;
+			IDL_tree_walk_real (&down_tfd, data);
+		}
 		break;
 
 	case IDLN_NATIVE:
@@ -1593,22 +1619,28 @@ static void IDL_tree_walk_real (IDL_tree_func_data *tfd, IDLTreeWalkRealData *da
 	case IDLN_TYPE_DCL:
 		down_tfd.tree = IDL_TYPE_DCL (p).type_spec;
 		IDL_tree_walk_real (&down_tfd, data);
-		down_tfd.tree = IDL_TYPE_DCL (p).dcls;
-		IDL_tree_walk_real (&down_tfd, data);
+		if ( (tfd->state->flags & IDL_WalkF_TypespecOnly)==0 ) {
+			down_tfd.tree = IDL_TYPE_DCL (p).dcls;
+			IDL_tree_walk_real (&down_tfd, data);
+		}
 		break;
 
 	case IDLN_CONST_DCL:
 		down_tfd.tree = IDL_CONST_DCL (p).const_type;
 		IDL_tree_walk_real (&down_tfd, data);
-		down_tfd.tree = IDL_CONST_DCL (p).ident;
-		IDL_tree_walk_real (&down_tfd, data);
-		down_tfd.tree = IDL_CONST_DCL (p).const_exp;
-		IDL_tree_walk_real (&down_tfd, data);
+		if ( (tfd->state->flags & IDL_WalkF_TypespecOnly)==0 ) {
+			down_tfd.tree = IDL_CONST_DCL (p).ident;
+			IDL_tree_walk_real (&down_tfd, data);
+			down_tfd.tree = IDL_CONST_DCL (p).const_exp;
+			IDL_tree_walk_real (&down_tfd, data);
+		}
 		break;
 
 	case IDLN_EXCEPT_DCL:
-		down_tfd.tree = IDL_EXCEPT_DCL (p).ident;
-		IDL_tree_walk_real (&down_tfd, data);
+		if ( (tfd->state->flags & IDL_WalkF_TypespecOnly)==0 ) {
+			down_tfd.tree = IDL_EXCEPT_DCL (p).ident;
+			IDL_tree_walk_real (&down_tfd, data);
+		}
 		down_tfd.tree = IDL_EXCEPT_DCL (p).members;
 		IDL_tree_walk_real (&down_tfd, data);
 		break;
@@ -1616,8 +1648,10 @@ static void IDL_tree_walk_real (IDL_tree_func_data *tfd, IDLTreeWalkRealData *da
 	case IDLN_ATTR_DCL:
 		down_tfd.tree = IDL_ATTR_DCL (p).param_type_spec;
 		IDL_tree_walk_real (&down_tfd, data);
-		down_tfd.tree = IDL_ATTR_DCL (p).simple_declarations;
-		IDL_tree_walk_real (&down_tfd, data);
+		if ( (tfd->state->flags & IDL_WalkF_TypespecOnly)==0 ) {
+			down_tfd.tree = IDL_ATTR_DCL (p).simple_declarations;
+			IDL_tree_walk_real (&down_tfd, data);
+		}
 		break;
 
 	case IDLN_OP_DCL:
@@ -1636,8 +1670,10 @@ static void IDL_tree_walk_real (IDL_tree_func_data *tfd, IDLTreeWalkRealData *da
 	case IDLN_PARAM_DCL:
 		down_tfd.tree = IDL_PARAM_DCL (p).param_type_spec;
 		IDL_tree_walk_real (&down_tfd, data);
-		down_tfd.tree = IDL_PARAM_DCL (p).simple_declarator;
-		IDL_tree_walk_real (&down_tfd, data);
+		if ( (tfd->state->flags & IDL_WalkF_TypespecOnly)==0 ) {
+			down_tfd.tree = IDL_PARAM_DCL (p).simple_declarator;
+			IDL_tree_walk_real (&down_tfd, data);
+		}
 		break;
 
 	case IDLN_FORWARD_DCL:
@@ -1646,41 +1682,53 @@ static void IDL_tree_walk_real (IDL_tree_func_data *tfd, IDLTreeWalkRealData *da
 		break;
 
 	case IDLN_TYPE_FIXED:
-		down_tfd.tree = IDL_TYPE_FIXED (p).positive_int_const;
-		IDL_tree_walk_real (&down_tfd, data);
-		down_tfd.tree = IDL_TYPE_FIXED (p).integer_lit;
-		IDL_tree_walk_real (&down_tfd, data);
+		if ( (tfd->state->flags & IDL_WalkF_TypespecOnly)==0 ) {
+			down_tfd.tree = IDL_TYPE_FIXED (p).positive_int_const;
+			IDL_tree_walk_real (&down_tfd, data);
+			down_tfd.tree = IDL_TYPE_FIXED (p).integer_lit;
+			IDL_tree_walk_real (&down_tfd, data);
+		}
 		break;
 
 	case IDLN_TYPE_STRING:
-		down_tfd.tree = IDL_TYPE_STRING (p).positive_int_const;
-		IDL_tree_walk_real (&down_tfd, data);
+		if ( (tfd->state->flags & IDL_WalkF_TypespecOnly)==0 ) {
+			down_tfd.tree = IDL_TYPE_STRING (p).positive_int_const;
+			IDL_tree_walk_real (&down_tfd, data);
+		}
 		break;
 
 	case IDLN_TYPE_WIDE_STRING:
-		down_tfd.tree = IDL_TYPE_WIDE_STRING (p).positive_int_const;
-		IDL_tree_walk_real (&down_tfd, data);
+		if ( (tfd->state->flags & IDL_WalkF_TypespecOnly)==0 ) {
+			down_tfd.tree = IDL_TYPE_WIDE_STRING (p).positive_int_const;
+			IDL_tree_walk_real (&down_tfd, data);
+		}
 		break;
 
 	case IDLN_TYPE_ENUM:
-		down_tfd.tree = IDL_TYPE_ENUM (p).ident;
-		IDL_tree_walk_real (&down_tfd, data);
-		down_tfd.tree = IDL_TYPE_ENUM (p).enumerator_list;
-		IDL_tree_walk_real (&down_tfd, data);
+		if ( (tfd->state->flags & IDL_WalkF_TypespecOnly)==0 ) {
+			down_tfd.tree = IDL_TYPE_ENUM (p).ident;
+			IDL_tree_walk_real (&down_tfd, data);
+			down_tfd.tree = IDL_TYPE_ENUM (p).enumerator_list;
+			IDL_tree_walk_real (&down_tfd, data);
+		}
 		break;
 
 	case IDLN_TYPE_SEQUENCE:
 		down_tfd.tree = IDL_TYPE_SEQUENCE (p).simple_type_spec;
 		IDL_tree_walk_real (&down_tfd, data);
-		down_tfd.tree = IDL_TYPE_SEQUENCE (p).positive_int_const;
-		IDL_tree_walk_real (&down_tfd, data);
+		if ( (tfd->state->flags & IDL_WalkF_TypespecOnly)==0 ) {
+			down_tfd.tree = IDL_TYPE_SEQUENCE (p).positive_int_const;
+			IDL_tree_walk_real (&down_tfd, data);
+		}
 		break;
 
 	case IDLN_TYPE_ARRAY:
 		down_tfd.tree = IDL_TYPE_ARRAY (p).ident;
 		IDL_tree_walk_real (&down_tfd, data);
-		down_tfd.tree = IDL_TYPE_ARRAY (p).size_list;
-		IDL_tree_walk_real (&down_tfd, data);
+		if ( (tfd->state->flags & IDL_WalkF_TypespecOnly)==0 ) {
+			down_tfd.tree = IDL_TYPE_ARRAY (p).size_list;
+			IDL_tree_walk_real (&down_tfd, data);
+		}
 		break;
 
 	case IDLN_TYPE_STRUCT:
@@ -1702,6 +1750,7 @@ static void IDL_tree_walk_real (IDL_tree_func_data *tfd, IDLTreeWalkRealData *da
 	case IDLN_CASE_STMT:
 		down_tfd.tree = IDL_CASE_STMT (p).labels;
 		IDL_tree_walk_real (&down_tfd, data);
+		/* FIXME */
 		down_tfd.tree = IDL_CASE_STMT (p).element_spec;
 		IDL_tree_walk_real (&down_tfd, data);
 		break;
@@ -1746,7 +1795,8 @@ static void IDL_tree_walk_real (IDL_tree_func_data *tfd, IDLTreeWalkRealData *da
 	tfd->state->bottom = tfd->up;
 }
 
-void IDL_tree_walk (IDL_tree p, IDL_tree_func_data *current,
+void IDL_tree_walk2 (IDL_tree p, IDL_tree_func_data *current,
+		    glong flags, 
 		    IDL_tree_func pre_tree_func, IDL_tree_func post_tree_func,
 		    gpointer user_data)
 {
@@ -1762,8 +1812,13 @@ void IDL_tree_walk (IDL_tree p, IDL_tree_func_data *current,
 
 	tfs.up = current ? current->state : NULL;
 	tfs.start = p;
+	tfs.flags = flags;
 
-	if (current) tfd = *current;
+	tfd.level = 0;
+	if (current) {
+	    tfd = *current;
+	    tfd.level = ((tfd.level / 1000)+1) * 1000;
+	}
 	tfd.state = &tfs;
 	tfd.up = current;
 	tfd.tree = p;
@@ -1771,9 +1826,17 @@ void IDL_tree_walk (IDL_tree p, IDL_tree_func_data *current,
 	IDL_tree_walk_real (&tfd, &data);
 }
 
+void IDL_tree_walk (IDL_tree p, IDL_tree_func_data *current,
+		    IDL_tree_func pre_tree_func, IDL_tree_func post_tree_func,
+		    gpointer user_data)
+{
+    IDL_tree_walk2 (p, current, /*flags*/0, 
+      pre_tree_func, post_tree_func, user_data);
+}
+
 void IDL_tree_walk_in_order (IDL_tree p, IDL_tree_func tree_func, gpointer user_data)
 {
-	IDL_tree_walk (p, NULL, tree_func, NULL, user_data);
+	IDL_tree_walk2 (p, NULL, /*flags*/0, tree_func, NULL, user_data);
 }
 
 static void __IDL_tree_free (IDL_tree p);
@@ -2586,7 +2649,7 @@ static void IDL_output_delim (IDL_tree p, IDL_tree_func_data *current,
 	delim.hit = FALSE;
 	delim.delim = str;
 
-	IDL_tree_walk (p, current,
+	IDL_tree_walk2 (p, current, /*flags*/0,
 		       (IDL_tree_func) IDL_output_delim_pre,
 		       (IDL_tree_func) IDL_output_delim_post,
 		       &delim);
@@ -2703,13 +2766,16 @@ static gboolean IDL_emit_IDL_ident_real (IDL_tree_func_data *tfd, IDL_output_dat
 
 	assert (IDL_NODE_TYPE (tfd->tree) == IDLN_IDENT);
 
-	if (!up_real || data->flags & IDLF_OUTPUT_NO_QUALIFY_IDENTS)
+	if (!up_real || data->flags & IDLF_OUTPUT_NO_QUALIFY_IDENTS) {
 		dataf (data, "%s", IDL_IDENT (tfd->tree).str);
-	else {
-		/* Determine minimal required levels of scoping */
-		assert (up_path != NULL);
-		scope = up_path->tree ? up_path->tree : up_real;
-		levels = IDL_ns_scope_levels_from_here (data->ns, tfd->tree, scope);
+	} else {
+		if ( up_path == 0 ) {
+		    levels = 0;
+		} else {
+		    /* Determine minimal required levels of scoping */
+		    scope = up_path->tree ? up_path->tree : up_real;
+		    levels = IDL_ns_scope_levels_from_here (data->ns, tfd->tree, scope);
+		}
 		s = IDL_ns_ident_to_qstring (IDL_IDENT_TO_NS (tfd->tree), "::", levels);
 		dataf (data, "%s", s);
 		g_free (s);
@@ -2735,7 +2801,7 @@ static gboolean IDL_emit_IDL_ident_force_pre (IDL_tree_func_data *tfd, IDL_outpu
 
 static gboolean IDL_emit_IDL_ident (IDL_tree ident, IDL_tree_func_data *tfd, IDL_output_data *data)
 {
-	IDL_tree_walk (ident, tfd,
+	IDL_tree_walk2 (ident, tfd, /*flags*/0,
 		       (IDL_tree_func) IDL_emit_IDL_ident_real, NULL,
 		       data);
 
@@ -2916,7 +2982,7 @@ static gboolean IDL_emit_IDL_type_pre (IDL_tree_func_data *tfd, IDL_output_data
 	case IDLN_TYPE_SEQUENCE:
 		dataf (data, "sequence<");
 		save_flag (idents, TRUE);
-		IDL_tree_walk (IDL_TYPE_SEQUENCE (tfd->tree).simple_type_spec, tfd,
+		IDL_tree_walk2 (IDL_TYPE_SEQUENCE (tfd->tree).simple_type_spec, tfd,/*flags*/0,
 			       (IDL_tree_func) IDL_emit_node_pre_func,
 			       (IDL_tree_func) IDL_emit_node_post_func,
 			       data);
@@ -2939,7 +3005,7 @@ static gboolean IDL_emit_IDL_type_pre (IDL_tree_func_data *tfd, IDL_output_data
 		IDL_emit_IDL_ident (IDL_TYPE_STRUCT (p).ident, tfd, data);
 		dataf (data, DELIM_SPACE);
 		IDL_emit_IDL_curly_brace_open (tfd, data);
-		IDL_tree_walk (IDL_TYPE_STRUCT (p).member_list, tfd,
+		IDL_tree_walk2 (IDL_TYPE_STRUCT (p).member_list, tfd,/*flags*/0,
 			       (IDL_tree_func) IDL_emit_node_pre_func,
 			       (IDL_tree_func) IDL_emit_node_post_func,
 			       data);
@@ -2961,13 +3027,13 @@ static gboolean IDL_emit_IDL_type_pre (IDL_tree_func_data *tfd, IDL_output_data
 		dataf (data, DELIM_SPACE);
 		dataf (data, "switch" DELIM_SPACE "(");
 		save_flag (idents, TRUE);
-		IDL_tree_walk (IDL_TYPE_UNION (p).switch_type_spec, tfd,
+		IDL_tree_walk2 (IDL_TYPE_UNION (p).switch_type_spec, tfd,/*flags*/0,
 			       (IDL_tree_func) IDL_emit_node_pre_func,
 			       (IDL_tree_func) IDL_emit_node_post_func,
 			       data);
 		restore_flag (idents);
 		dataf (data, ")" DELIM_SPACE "{"); nl ();
-		IDL_tree_walk (IDL_TYPE_UNION (p).switch_body, tfd,
+		IDL_tree_walk2 (IDL_TYPE_UNION (p).switch_body, tfd,/*flags*/0,
 			       (IDL_tree_func) IDL_emit_node_pre_func,
 			       (IDL_tree_func) IDL_emit_node_post_func,
 			       data);
@@ -3045,7 +3111,7 @@ static gboolean IDL_emit_IDL_attr_dcl_pre (IDL_tree_func_data *tfd, IDL_output_d
 	if (IDL_ATTR_DCL (tfd->tree).f_readonly) dataf (data, "readonly" DELIM_SPACE);
 	dataf (data, "attribute" DELIM_SPACE);
 	save_flag (idents, TRUE);
-	IDL_tree_walk (IDL_ATTR_DCL (tfd->tree).param_type_spec, tfd,
+	IDL_tree_walk2 (IDL_ATTR_DCL (tfd->tree).param_type_spec, tfd,/*flags*/0,
 		       (IDL_tree_func) IDL_emit_node_pre_func,
 		       (IDL_tree_func) IDL_emit_node_post_func,
 		       data);
@@ -3067,7 +3133,7 @@ static gboolean IDL_emit_IDL_op_dcl_pre (IDL_tree_func_data *tfd, IDL_output_dat
 	if (IDL_OP_DCL (tfd->tree).f_oneway) dataf (data, "oneway" DELIM_SPACE);
 	if (IDL_OP_DCL (tfd->tree).op_type_spec) {
 		save_flag (idents, TRUE);
-		IDL_tree_walk (IDL_OP_DCL (tfd->tree).op_type_spec, tfd,
+		IDL_tree_walk2 (IDL_OP_DCL (tfd->tree).op_type_spec, tfd,/*flags*/0,
 			       (IDL_tree_func) IDL_emit_node_pre_func,
 			       (IDL_tree_func) IDL_emit_node_post_func,
 			       data);
@@ -3117,7 +3183,7 @@ static gboolean IDL_emit_IDL_param_dcl_pre (IDL_tree_func_data *tfd, IDL_output_
 	case IDL_PARAM_INOUT: dataf (data, "inout" DELIM_SPACE); break;
 	}
 	save_flag (idents, TRUE);
-	IDL_tree_walk (IDL_PARAM_DCL (tfd->tree).param_type_spec, tfd,
+	IDL_tree_walk2 (IDL_PARAM_DCL (tfd->tree).param_type_spec, tfd,/*flags*/0,
 		       (IDL_tree_func) IDL_emit_node_pre_func,
 		       (IDL_tree_func) IDL_emit_node_post_func,
 		       data);
@@ -3139,7 +3205,7 @@ static gboolean IDL_emit_IDL_type_dcl_pre (IDL_tree_func_data *tfd, IDL_output_d
 	dataf (data, "typedef" DELIM_SPACE);
 	save_flag (idents, TRUE);
 	save_flag (su_def, TRUE);
-	IDL_tree_walk (IDL_TYPE_DCL (tfd->tree).type_spec, tfd,
+	IDL_tree_walk2 (IDL_TYPE_DCL (tfd->tree).type_spec, tfd,/*flags*/0,
 		       (IDL_tree_func) IDL_emit_node_pre_func,
 		       (IDL_tree_func) IDL_emit_node_post_func,
 		       data);
@@ -3148,7 +3214,7 @@ static gboolean IDL_emit_IDL_type_dcl_pre (IDL_tree_func_data *tfd, IDL_output_d
 	down_tfd.up = tfd;
 	for (q = IDL_TYPE_DCL (tfd->tree).dcls; q; q = IDL_LIST (q).next) {
 		down_tfd.tree = q;
-		IDL_tree_walk (IDL_LIST (q).data, &down_tfd,
+		IDL_tree_walk2 (IDL_LIST (q).data, &down_tfd,/*flags*/0,
 			       (IDL_tree_func) IDL_emit_node_pre_func,
 			       (IDL_tree_func) IDL_emit_node_post_func,
 			       data);
@@ -3166,7 +3232,7 @@ static gboolean IDL_emit_IDL_const_dcl_pre (IDL_tree_func_data *tfd, IDL_output_
 {
 	idataf (data, "const" DELIM_SPACE);
 	save_flag (idents, TRUE);
-	IDL_tree_walk (IDL_CONST_DCL (tfd->tree).const_type, tfd,
+	IDL_tree_walk2 (IDL_CONST_DCL (tfd->tree).const_type, tfd,/*flags*/0,
 		       (IDL_tree_func) IDL_emit_node_pre_func,
 		       (IDL_tree_func) IDL_emit_node_post_func,
 		       data);
@@ -3175,7 +3241,7 @@ static gboolean IDL_emit_IDL_const_dcl_pre (IDL_tree_func_data *tfd, IDL_output_
 	IDL_emit_IDL_ident (IDL_CONST_DCL (tfd->tree).ident, tfd, data);
 	dataf (data, DELIM_SPACE "=" DELIM_SPACE);
 	save_flag (literals, TRUE);
-	IDL_tree_walk (IDL_CONST_DCL (tfd->tree).const_exp, tfd,
+	IDL_tree_walk2 (IDL_CONST_DCL (tfd->tree).const_exp, tfd,/*flags*/0,
 		       (IDL_tree_func) IDL_emit_node_pre_func,
 		       (IDL_tree_func) IDL_emit_node_post_func,
 		       data);
@@ -3222,7 +3288,7 @@ static gboolean IDL_emit_IDL_case_stmt_pre (IDL_tree_func_data *tfd, IDL_output_
 		if (IDL_LIST (q).data) {
 			down_tfd.tree = q;
 			idataf (data, "case" DELIM_SPACE);
-			IDL_tree_walk (IDL_LIST (q).data, &down_tfd,
+			IDL_tree_walk2 (IDL_LIST (q).data, &down_tfd,/*flags*/0,
 				       (IDL_tree_func) IDL_emit_node_pre_func,
 				       (IDL_tree_func) IDL_emit_node_post_func,
 				       data);
@@ -3240,7 +3306,7 @@ static gboolean IDL_emit_IDL_case_stmt_pre (IDL_tree_func_data *tfd, IDL_output_
 
 static gboolean IDL_emit_IDL_case_stmt_post (IDL_tree_func_data *tfd, IDL_output_data *data)
 {
-	IDL_tree_walk (IDL_CASE_STMT (tfd->tree).element_spec, tfd,
+	IDL_tree_walk2 (IDL_CASE_STMT (tfd->tree).element_spec, tfd, /*flags*/0,
 		       (IDL_tree_func) IDL_emit_node_pre_func,
 		       (IDL_tree_func) IDL_emit_node_post_func,
 		       data);
@@ -3253,7 +3319,7 @@ static gboolean IDL_emit_IDL_member_pre (IDL_tree_func_data *tfd, IDL_output_dat
 {
 	IDL_emit_IDL_indent (tfd, data);
 	save_flag (idents, TRUE);
-	IDL_tree_walk (IDL_MEMBER (tfd->tree).type_spec, tfd,
+	IDL_tree_walk2 (IDL_MEMBER (tfd->tree).type_spec, tfd, /*flags*/0,
 		       (IDL_tree_func) IDL_emit_node_pre_func,
 		       (IDL_tree_func) IDL_emit_node_post_func,
 		       data);
@@ -3421,7 +3487,7 @@ void IDL_tree_to_IDL (IDL_tree p, IDL_ns ns, FILE *output, unsigned long output_
 	if (ns == NULL)
 		data.flags |= IDLF_OUTPUT_NO_QUALIFY_IDENTS;
 
-	IDL_tree_walk (p, NULL,
+	IDL_tree_walk2 (p, NULL, /*flags*/0,
 		       (IDL_tree_func) IDL_emit_node_pre_func,
 		       (IDL_tree_func) IDL_emit_node_post_func,
 		       &data);
@@ -3444,7 +3510,7 @@ GString *IDL_tree_to_IDL_string (IDL_tree p, IDL_ns ns, unsigned long output_fla
 	if (ns == NULL)
 		data.flags |= IDLF_OUTPUT_NO_QUALIFY_IDENTS;
 
-	IDL_tree_walk (p, NULL,
+	IDL_tree_walk2 (p, NULL, /*flags*/0,
 		       (IDL_tree_func) IDL_emit_node_pre_func,
 		       (IDL_tree_func) IDL_emit_node_post_func,
 		       &data);
@@ -3452,6 +3518,160 @@ GString *IDL_tree_to_IDL_string (IDL_tree p, IDL_ns ns, unsigned long output_fla
 	return data.u.s;
 }
 
+typedef struct {
+    IDL_tree	searchNode;
+    gboolean	found;
+} ContainsNodeWalkerInfo;
+
+static gboolean 
+contains_node_walker (IDL_tree_func_data *tfd, gpointer data) {
+	ContainsNodeWalkerInfo	*info = data;
+	/* skip root! */
+	if ( tfd->up!=0 && tfd->tree == info->searchNode ) {
+		info->found = TRUE;
+		return FALSE;
+	}
+ 	return TRUE;	/* continue walking */
+}
+
+gboolean IDL_tree_contains_node(IDL_tree tree, IDL_tree node) {
+	ContainsNodeWalkerInfo	info;
+	info.searchNode = node;
+	info.found = 0;
+	IDL_tree_walk2 (tree, /*curwalk*/NULL, /*flags*/0,
+	  /*pre*/contains_node_walker, /*post*/NULL, &info);
+	return info.found;
+}
+
+struct IDL_recursive_walker_info {
+	GSList   *ident_list;
+	gboolean  recursive;
+	};
+
+/*
+ * IDL_tree_is_recursive_walker_pre:
+ * @walk_frame: a #IDL_tree_func_data structure
+ * @user_data: a #IDL_recursive_walker_info structure.
+ *
+ * If the current node in the parse tree is a structure or union
+ * add its IDLN_IDENT to the ident list contained in @user_data;
+ *
+ * If the current node in the parse tree is a sequence, check to
+ * see if the sequence element type is on the list of IDLN_IDENTs.
+ *
+ * Return value: %FALSE if recursion is detected - this stops the
+ *               walk to the parse tree -, %TRUE otherwise.
+ */
+static gboolean
+IDL_tree_is_recursive_walker_pre (IDL_tree_func_data *walk_frame,
+				  gpointer            user_data)
+{
+	struct IDL_recursive_walker_info *info = user_data;
+	IDL_tree                          node = walk_frame->tree;
+
+	switch (IDL_NODE_TYPE (node)) {
+	case IDLN_TYPE_STRUCT:
+	case IDLN_TYPE_UNION:
+		info->ident_list = g_slist_prepend (info->ident_list,
+						    IDL_TYPE_STRUCT (node).ident);
+		break;
+	case IDLN_TYPE_SEQUENCE: {
+		IDL_tree  seq_type = IDL_TYPE_SEQUENCE (node).simple_type_spec;
+		GSList   *l = info->ident_list;
+
+		if (IDL_NODE_TYPE (seq_type) != IDLN_IDENT)
+			break;
+
+		g_assert (IDL_IDENT (seq_type).repo_id);
+
+		for (; l; l = l->next) {
+			IDL_tree n = (IDL_tree)l->data;
+
+			g_assert (IDL_IDENT (n).repo_id);
+
+			if (!strcmp (IDL_IDENT (n).repo_id,
+				     IDL_IDENT (seq_type).repo_id)) {
+
+				info->recursive = TRUE;
+				return FALSE;
+			}
+		}
+		}
+		break;
+	default:
+		break;
+	}
+
+	return TRUE;
+}
+
+/*
+ * IDL_tree_is_recursive_walker_post:
+ * @walk_frame: a #IDL_tree_func_data structure
+ * @user_data: a #IDL_recursive_walker_info structure.
+ *
+ * If the current node in the parse tree is a structure or
+ * union, remove the IDLN_IDENT from the ident list.
+ *
+ * Return value: %TRUE.
+ */
+static gboolean
+IDL_tree_is_recursive_walker_post (IDL_tree_func_data *walk_frame,
+				   gpointer            user_data)
+{
+	struct IDL_recursive_walker_info *info = user_data;
+	IDL_tree                          node = walk_frame->tree;
+
+	switch (IDL_NODE_TYPE (node)) {
+	case IDLN_TYPE_STRUCT:
+	case IDLN_TYPE_UNION: {
+		GSList *link = info->ident_list;
+
+		g_assert (((IDL_tree)link->data) == IDL_TYPE_STRUCT (node).ident);
+
+		info->ident_list = g_slist_remove_link (info->ident_list,
+						        link);
+		g_slist_free_1 (link);
+		}
+		break;
+	default:
+		break;
+	}
+
+	return TRUE;
+}
+
+/*
+ * IDL_tree_is_recursive:
+ * @tree: a #IDL_tree.
+ * @hasRecur: dummy argument, not used, deprecated, compatibility cruft.
+ *
+ * Checks whether or not @tree contains recursive types.
+ *
+ * Recursive types may only be constructed using structs/unions which
+ * contain sequences that hold the parent struct/union type.
+ *
+ * Return value: %TRUE if @tree contains recursive types,
+ *               %FALSE otherwise.
+ */
+gboolean
+IDL_tree_is_recursive (IDL_tree tree, gpointer dummy)
+{
+	struct IDL_recursive_walker_info info;
+
+	info.ident_list = NULL;
+	info.recursive  = FALSE;
+
+	IDL_tree_walk2 (tree, 0, IDL_WalkF_TypespecOnly, 
+			IDL_tree_is_recursive_walker_pre, 
+			IDL_tree_is_recursive_walker_post, 
+			&info);
+
+	g_assert (!info.ident_list);
+
+	return info.recursive;
+}
+
 /*
  * Local variables:
  * mode: C
